import asyncio
import io
import logging
import random
import shelve
from collections import OrderedDict, defaultdict
from dataclasses import dataclass
from mimetypes import guess_type
from pathlib import Path
from typing import Optional, Union

import aiohttp
import maufbapi.types.graphql
from maufbapi import AndroidAPI, AndroidMQTT, AndroidState
from maufbapi.types import mqtt as mqtt_t
from maufbapi.types.graphql import Participant, ParticipantNode, Thread
from maufbapi.types.graphql.responses import FriendshipStatus
from slixmpp.exceptions import XMPPError

from slidge import *


class Gateway(BaseGateway):
    REGISTRATION_INSTRUCTIONS = "Enter facebook credentials"
    REGISTRATION_FIELDS = [
        FormField(var="email", label="Email", required=True),
        FormField(var="password", label="Password", required=True, private=True),
    ]

    ROSTER_GROUP = "Facebook"

    COMPONENT_NAME = "Facebook (slidge)"
    COMPONENT_TYPE = "facebook"
    COMPONENT_AVATAR = "https://upload.wikimedia.org/wikipedia/commons/thumb/6/6c/Facebook_Messenger_logo_2018.svg/480px-Facebook_Messenger_logo_2018.svg.png"

    SEARCH_TITLE = "Search in your facebook friends"
    SEARCH_INSTRUCTIONS = "Enter something that can be used to search for one of your friends, eg, a first name"
    SEARCH_FIELDS = [FormField(var="query", label="Term(s)")]


class Contact(LegacyContact["Session"]):
    legacy_id: str  # facebook username, as in facebook.com/name.surname123

    def __init__(self, *a, **k):
        super(Contact, self).__init__(*a, **k)
        self._fb_id: Optional[int] = None

    async def fb_id(self):
        if self._fb_id is None:
            results = await self.session.api.search(
                self.legacy_id, entity_types=["user"]
            )
            for search_result in results.search_results.edges:
                result = search_result.node
                if (
                    isinstance(result, Participant)
                    and result.username == self.legacy_id
                ):
                    self._fb_id = int(result.id)
                    break
            else:
                raise XMPPError(
                    "not-found", text=f"Cannot find the facebook ID of {self.legacy_id}"
                )
            self.session.contacts.by_fb_id_dict[self._fb_id] = self
        return self._fb_id

    async def populate_from_participant(
        self, participant: ParticipantNode, update_avatar=True
    ):
        if self.legacy_id != participant.messaging_actor.username:
            raise RuntimeError(
                "Attempted to populate a contact with a non-corresponding participant"
            )
        self.name = participant.messaging_actor.name
        self._fb_id = int(participant.id)
        if self.avatar is None or update_avatar:
            async with aiohttp.ClientSession() as session:
                async with session.get(
                    participant.messaging_actor.profile_pic_large.uri
                ) as response:
                    response.raise_for_status()
                    self.avatar = await response.read()


class Roster(LegacyRoster[Contact, "Session"]):
    def __init__(self, *a, **k):
        super().__init__(*a, **k)
        self.by_fb_id_dict: dict[int, Contact] = {}

    async def by_fb_id(self, fb_id: int) -> "Contact":
        contact = self.by_fb_id_dict.get(fb_id)
        if contact is None:
            thread = (await self.session.api.fetch_thread_info(fb_id))[0]
            return await self.by_thread(thread)
        return contact

    async def by_thread_key(self, t: mqtt_t.ThreadKey):
        if is_group_thread(t):
            raise ValueError("Thread seems to be a group thread")
        return await self.by_fb_id(t.other_user_id)

    async def by_thread(self, t: Thread):
        if t.is_group_thread:
            raise RuntimeError("Tried to populate a user from a group chat")

        if len(t.all_participants.nodes) != 2:
            raise RuntimeError(
                "Tried is not a group chat but doesn't have 2 participants ‽"
            )

        for participant in t.all_participants.nodes:
            if participant.id != self.session.me.id:
                break
        else:
            raise RuntimeError(
                "Couldn't find friend in thread participants", t.all_participants
            )

        contact = self.by_legacy_id(participant.messaging_actor.username)
        await contact.populate_from_participant(participant)
        self.by_fb_id_dict[int(participant.id)] = contact
        return contact


class Session(BaseSession[Contact, Roster, Gateway]):
    fb_state: AndroidState

    shelf_path: Path
    mqtt: AndroidMQTT
    api: AndroidAPI

    me: maufbapi.types.graphql.OwnInfo

    sent_messages: defaultdict[int, "Messages"]
    received_messages: defaultdict[int, "Messages"]
    # keys = "contact ID"

    ack_futures: dict[int, asyncio.Future["FacebookMessage"]]
    # keys = "offline thread ID"
    reaction_futures: dict[str, asyncio.Future[None]]
    unsend_futures: dict[str, asyncio.Future[None]]
    # keys = "facebook message id"

    contacts: Roster

    def post_init(self):
        self.shelf_path = self.xmpp.home_dir / self.user.bare_jid
        self.ack_futures = {}
        self.reaction_futures: dict[str, asyncio.Future] = {}
        self.unsend_futures: dict[str, asyncio.Future] = {}
        self.sent_messages = defaultdict(Messages)
        self.received_messages = defaultdict(Messages)

    async def login(self):
        shelf: shelve.Shelf[AndroidState]
        with shelve.open(str(self.shelf_path)) as shelf:
            try:
                self.fb_state = s = shelf["state"]
            except KeyError:
                s = AndroidState()
                self.api = api = AndroidAPI(state=s)
                s.generate(random.randbytes(30))  # type: ignore
                await api.mobile_config_sessionless()
                try:
                    login = await api.login(
                        email=self.user.registration_form["email"],
                        password=self.user.registration_form["password"],
                    )
                except maufbapi.http.errors.IncorrectPassword:
                    self.send_gateway_message("Incorrect password")
                    raise
                except maufbapi.http.errors.TwoFactorRequired:
                    code = await self.input(
                        "Reply to this message with your 2 factor authentication code"
                    )
                    login = await api.login_2fa(
                        email=self.user.registration_form["email"], code=code
                    )
                log.debug(login)
                self.fb_state = shelf["state"] = api.state
            else:
                self.api = api = AndroidAPI(state=s)
        self.mqtt = AndroidMQTT(api.state)
        self.me = await self.api.get_self()
        self.me.id = int(self.me.id)  # bug in maufbapi?
        await self.add_friends()
        self.mqtt.seq_id_update_callback = lambda i: setattr(self.mqtt, "seq_id", i)
        self.mqtt.add_event_handler(mqtt_t.Message, self.on_fb_message)
        self.mqtt.add_event_handler(mqtt_t.ExtendedMessage, self.on_fb_message)
        self.mqtt.add_event_handler(mqtt_t.ReadReceipt, self.on_fb_message_read)
        self.mqtt.add_event_handler(mqtt_t.TypingNotification, self.on_fb_typing)
        self.mqtt.add_event_handler(mqtt_t.OwnReadReceipt, self.on_fb_user_read)
        self.mqtt.add_event_handler(mqtt_t.Reaction, self.on_fb_reaction)
        self.mqtt.add_event_handler(mqtt_t.UnsendMessage, self.on_fb_unsend)

        self.mqtt.add_event_handler(mqtt_t.NameChange, self.on_fb_event)
        self.mqtt.add_event_handler(mqtt_t.AvatarChange, self.on_fb_event)
        self.mqtt.add_event_handler(mqtt_t.Presence, self.on_fb_event)
        self.mqtt.add_event_handler(mqtt_t.AddMember, self.on_fb_event)
        self.mqtt.add_event_handler(mqtt_t.RemoveMember, self.on_fb_event)
        self.mqtt.add_event_handler(mqtt_t.ThreadChange, self.on_fb_event)
        self.mqtt.add_event_handler(mqtt_t.MessageSyncError, self.on_fb_event)
        self.mqtt.add_event_handler(mqtt_t.ForcedFetch, self.on_fb_event)
        # self.mqtt.add_event_handler(Connect, self.on_connect)
        # self.mqtt.add_event_handler(Disconnect, self.on_disconnect)
        self.xmpp.loop.create_task(self.mqtt.listen(self.mqtt.seq_id))
        return f"Connected as '{self.me.name} <{self.me.email}>'"

    async def add_friends(self, n=2):
        thread_list = await self.api.fetch_thread_list(msg_count=0, thread_count=n)
        self.mqtt.seq_id = int(thread_list.sync_sequence_id)
        log.debug("SEQ ID: %s", self.mqtt.seq_id)
        self.log.debug("Thread list: %s", thread_list)
        self.log.debug("Thread list page info: %s", thread_list.page_info)
        for t in thread_list.nodes:
            if t.is_group_thread:
                log.debug("Skipping group: %s", t)
                continue
            c = await self.contacts.by_thread(t)
            await c.add_to_roster()
            c.online()

    async def logout(self):
        pass

    async def send_text(self, t: str, c: Contact, *, reply_to_msg_id=None) -> str:
        resp: mqtt_t.SendMessageResponse = await self.mqtt.send_message(
            target=(fb_id := await c.fb_id()),
            message=t,
            is_group=False,
            reply_to=reply_to_msg_id,
        )
        fut = self.ack_futures[
            resp.offline_threading_id
        ] = self.xmpp.loop.create_future()
        log.debug("Send message response: %s", resp)
        if not resp.success:
            raise XMPPError(resp.error_message)
        fb_msg = await fut
        self.sent_messages[fb_id].add(fb_msg)
        return fb_msg.mid

    async def send_file(self, u: str, c: Contact, *, reply_to_msg_id=None):
        async with aiohttp.ClientSession() as s:
            async with s.get(u) as r:
                data = await r.read()
        oti = self.mqtt.generate_offline_threading_id()
        fut = self.ack_futures[oti] = self.xmpp.loop.create_future()
        resp = await self.api.send_media(
            data=data,
            file_name=u.split("/")[-1],
            mimetype=guess_type(u)[0] or "application/octet-stream",
            offline_threading_id=oti,
            chat_id=await c.fb_id(),
            is_group=False,
            reply_to=reply_to_msg_id,
        )
        ack = await fut
        log.debug("Upload ack: %s", ack)
        return resp.media_id

    async def active(self, c: Contact):
        pass

    async def inactive(self, c: Contact):
        pass

    async def composing(self, c: Contact):
        await self.mqtt.set_typing(target=await c.fb_id())

    async def paused(self, c: Contact):
        await self.mqtt.set_typing(target=await c.fb_id(), typing=False)

    async def displayed(self, legacy_msg_id: str, c: Contact):
        fb_id = await c.fb_id()
        try:
            t = self.received_messages[fb_id].by_mid[legacy_msg_id].timestamp_ms
        except KeyError:
            log.debug("Cannot find the timestamp of %s", legacy_msg_id)
        else:
            await self.mqtt.mark_read(target=fb_id, read_to=t, is_group=False)

    async def on_fb_message(self, evt: Union[mqtt_t.Message, mqtt_t.ExtendedMessage]):
        if isinstance(evt, mqtt_t.ExtendedMessage):
            reply_to = evt.reply_to_message.metadata.id
            msg = evt.message
        else:
            reply_to = None
            msg = evt
        meta = msg.metadata
        if is_group_thread(thread_key := meta.thread):
            return
        contact = await self.contacts.by_thread_key(thread_key)

        if not contact.added_to_roster:
            await contact.add_to_roster()

        log.debug("Facebook message: %s", evt)
        fb_msg = FacebookMessage(mid=meta.id, timestamp_ms=meta.timestamp)
        if meta.sender == self.me.id:
            try:
                fut = self.ack_futures.pop(meta.offline_threading_id)
            except KeyError:
                log.debug("Received carbon %s - %s", meta.id, msg.text)
                contact.carbon(body=msg.text, legacy_id=meta.id)
                log.debug("Sent carbon")
                self.sent_messages[thread_key.other_user_id].add(fb_msg)
            else:
                log.debug("Received echo of %s", meta.offline_threading_id)
                fut.set_result(fb_msg)
        else:
            if msg.text:
                contact.send_text(
                    msg.text, legacy_msg_id=meta.id, reply_to_msg_id=reply_to
                )
            if msg.attachments:
                async with aiohttp.ClientSession() as c:
                    for a in msg.attachments:
                        url = (
                            ((v := a.video_info) and v.download_url)
                            or ((au := a.audio_info) and au.url)
                            or a.image_info.uri_map.get(0)
                        )
                        if url is None:
                            continue
                        async with c.get(url) as r:
                            await contact.send_file(
                                filename=a.file_name,
                                content_type=a.mime_type,
                                input_file=io.BytesIO(await r.read()),
                            )
            self.received_messages[thread_key.other_user_id].add(fb_msg)

    async def on_fb_message_read(self, receipt: mqtt_t.ReadReceipt):
        log.debug("Facebook read: %s", receipt)
        try:
            mid = self.sent_messages[receipt.user_id].pop_up_to(receipt.read_to).mid
        except KeyError:
            log.debug("Cannot find MID of %s", receipt.read_to)
        else:
            contact = await self.contacts.by_thread_key(receipt.thread)
            contact.displayed(mid)

    async def on_fb_typing(self, notification: mqtt_t.TypingNotification):
        log.debug("Facebook typing: %s", notification)
        c = await self.contacts.by_fb_id(notification.user_id)
        if notification.typing_status:
            c.composing()
        else:
            c.paused()

    async def on_fb_user_read(self, receipt: mqtt_t.OwnReadReceipt):
        log.debug("Facebook own read: %s", receipt)
        when = receipt.read_to
        for thread in receipt.threads:
            c = await self.contacts.by_fb_id(thread.other_user_id)
            try:
                mid = self.received_messages[await c.fb_id()].pop_up_to(when).mid
            except KeyError:
                log.debug("Cannot find mid of %s", when)
                continue
            c.carbon_read(mid)

    async def on_fb_reaction(self, reaction: mqtt_t.Reaction):
        self.log.debug("Reaction: %s", reaction)
        if is_group_thread(tk := reaction.thread):
            return
        contact = await self.contacts.by_thread_key(tk)
        mid = reaction.message_id
        if reaction.reaction_sender_id == self.me.id:
            try:
                f = self.reaction_futures.pop(mid)
            except KeyError:
                contact.carbon_react(mid, reaction.reaction or "")
            else:
                f.set_result(None)
        else:
            contact.react(reaction.message_id, reaction.reaction or "")

    async def on_fb_unsend(self, unsend: mqtt_t.UnsendMessage):
        self.log.debug("Unsend: %s", unsend)
        if is_group_thread(tk := unsend.thread):
            return
        contact = await self.contacts.by_thread_key(tk)
        mid = unsend.message_id
        if unsend.user_id == self.me.id:
            try:
                f = self.unsend_futures.pop(mid)
            except KeyError:
                contact.carbon_retract(mid)
            else:
                f.set_result(None)
        else:
            contact.retract(unsend.message_id)

    async def correct(self, text: str, legacy_msg_id: str, c: Contact):
        await self.api.unsend(legacy_msg_id)
        return await self.send_text(text, c)

    async def react(self, legacy_msg_id: str, emojis: list[str], c: Contact):
        if len(emojis) == 0:
            emoji = None
        else:
            emoji = emojis[-1]
            if len(emojis) > 1:  # only reaction per msg on facebook
                c.carbon_react(legacy_msg_id, emoji)
        f = self.reaction_futures[legacy_msg_id] = self.xmpp.loop.create_future()
        await self.api.react(legacy_msg_id, emoji)
        await f

    async def retract(self, legacy_msg_id: str, c: Contact):
        f = self.unsend_futures[legacy_msg_id] = self.xmpp.loop.create_future()
        await self.api.unsend(legacy_msg_id)
        await f

    async def search(self, form_values: dict[str, str]) -> SearchResult:
        results = await self.api.search(form_values["query"], entity_types=["user"])
        log.debug("Search results: %s", results)
        items = []
        for search_result in results.search_results.edges:
            result = search_result.node
            if isinstance(result, Participant):
                is_friend = (
                    friend := result.friendship_status
                ) is not None and friend == FriendshipStatus.ARE_FRIENDS
                items.append(
                    {
                        "name": result.name + " (friend)"
                        if is_friend
                        else " (not friend)",
                        "jid": f"{result.username}@{self.xmpp.boundjid.bare}",
                    }
                )

        return SearchResult(
            fields=[
                FormField(var="name", label="Name"),
                FormField(var="jid", label="JID", type="jid-single"),
            ],
            items=items,
        )

    @staticmethod
    async def on_fb_event(evt):
        log.debug("Facebook event: %s", evt)


@dataclass
class FacebookMessage:
    mid: str
    timestamp_ms: int


class Messages:
    def __init__(self):
        self.by_mid: OrderedDict[str, FacebookMessage] = OrderedDict()
        self.by_timestamp_ms: OrderedDict[int, FacebookMessage] = OrderedDict()

    def __len__(self):
        return len(self.by_mid)

    def add(self, m: FacebookMessage):
        self.by_mid[m.mid] = m
        self.by_timestamp_ms[m.timestamp_ms] = m

    def pop_up_to(self, approx_t: int) -> FacebookMessage:
        i = 0
        for i, t in enumerate(self.by_timestamp_ms.keys()):
            if t > approx_t:
                i -= 1
                break
        for j, t in enumerate(list(self.by_timestamp_ms.keys())):
            msg = self.by_timestamp_ms.pop(t)
            self.by_mid.pop(msg.mid)
            if j == i:
                return msg
        else:
            raise KeyError(approx_t)


def is_group_thread(t: mqtt_t.ThreadKey):
    return t.other_user_id is None and t.thread_fbid is not None


log = logging.getLogger(__name__)



from typing import TYPE_CHECKING, Union

import discord as di

if TYPE_CHECKING:
    from .session import Session


class Discord(di.Client):
    def __init__(self, session: "Session"):
        super().__init__()
        self.session = session

    async def on_ready(self):
        self.session.ready_future.set_result(True)
        print(f"Logged on as {self.user}")

    async def on_message(self, message: di.Message):
        if (author := message.author) == self.user:
            async with self.session.send_lock:
                fut = self.session.send_futures.get(message.id)
            if fut is None:
                channel = message.channel
                if isinstance(channel, di.DMChannel):
                    self.session.contacts.by_discord_user(channel.recipient).carbon(
                        message.content
                    )
                else:
                    self.session.log.debug("Ignoring group chat carbon")
            else:
                fut.set_result(True)
        else:
            contact = self.session.contacts.by_discord_user(author)
            reply_to = message.reference.message_id if message.reference else None
            if content := message.content:
                contact.send_text(
                    content,
                    legacy_msg_id=message.id,
                    reply_to_msg_id=reply_to,
                )
            for attachment in message.attachments:
                await contact.send_file(
                    url=attachment.url,
                    filename=attachment.filename,
                    content_type=attachment.content_type,
                    reply_to_msg_id=reply_to,
                )

    async def on_typing(self, channel, user, _when):
        if user != self.user and isinstance(channel, di.DMChannel):
            self.session.contacts.by_discord_user(user).composing()

    async def on_message_edit(self, before: di.Message, after: di.Message):
        if not isinstance(after.channel, di.DMChannel):
            return
        if before.content == after.content:
            return
        if (author := after.author) == self.user:
            fut = self.session.edit_futures.get(after.id)
            if fut is None:
                self.session.contacts.by_discord_user(
                    after.channel.recipient
                ).carbon_correct(after.id, after.content)
            else:
                fut.set_result(True)
        else:
            self.session.contacts.by_discord_user(author).correct(
                after.id, after.content
            )

    async def on_message_delete(self, m: di.Message):
        if not isinstance(m.channel, di.DMChannel):
            return
        if (author := m.author) == self.user:
            fut = self.session.delete_futures.get(m.id)
            if fut is None:
                self.session.contacts.by_discord_user(
                    m.channel.recipient
                ).carbon_retract(m.id)
            else:
                fut.set_result(True)
        else:
            self.session.contacts.by_discord_user(author).retract(m.id)

    async def on_reaction_add(
        self, reaction: di.Reaction, user: Union[di.User, di.ClientUser]
    ):
        await self.update_reactions(reaction, user)

    async def on_reaction_remove(
        self, reaction: di.Reaction, user: Union[di.User, di.ClientUser]
    ):
        await self.update_reactions(reaction, user)

    async def update_reactions(
        self, reaction: di.Reaction, user: Union[di.User, di.ClientUser]
    ):
        message: di.Message = reaction.message
        if not isinstance(message.channel, di.DMChannel):
            return

        if user == self.user:
            self.session.update_reactions(message)
        else:
            await self.session.contacts.by_discord_user(user).update_reactions(message)



"""
Just a stab at steam chat.

Unfortunately the library underneath uses gevent and there is probably some work to do
to make it play nice with asyncio.

Right now, listing friends + send them messages works BUT in a blocking way.

Asyncsteampy https://github.com/somespecialone/asyncsteampy
might be interesting as it uses python's asyncio BUT the
login process seem a little too exotic for my taste.
"""
import asyncio
from typing import Any

import steam.enums
from steam.client import SteamClient
from steam.client.user import SteamUser
from steam.core.msg import MsgProto
from steam.enums.common import EPersonaState, EResult
from steam.enums.emsg import EMsg
from steam.protobufs.steammessages_friendmessages_pb2 import (
    k_EMessageReactionType_Emoticon,
)
from steam.steamid import SteamID

from slidge import *
from slidge.util import BiDict


class Gateway(BaseGateway["Session"]):
    REGISTRATION_INSTRUCTIONS = "Enter steam credentials"
    REGISTRATION_FIELDS = [
        FormField(var="username", label="Steam username", required=True),
        FormField(var="password", label="Password", private=True, required=True),
    ]

    ROSTER_GROUP = "Steam"

    COMPONENT_NAME = "Steam (slidge)"
    COMPONENT_TYPE = "steam"

    COMPONENT_AVATAR = "https://logos-download.com/wp-content/uploads/2016/05/Steam_icon_logo_logotype.png"


class Contact(LegacyContact["Session"]):
    MARKS = False
    CORRECTION = False
    RETRACTION = False

    def update_status(self, persona_state: EPersonaState):
        if persona_state == EPersonaState.Offline:
            self.offline()
        elif persona_state == EPersonaState.Online:
            self.online()
        elif persona_state == EPersonaState.Busy:
            self.busy()
        elif persona_state == EPersonaState.Away:
            self.away()


class Roster(LegacyRoster[Contact, "Session"]):
    @staticmethod
    def jid_username_to_legacy_id(jid_username: str) -> int:
        return int(jid_username)


class Session(BaseSession[Contact, Roster, Gateway]):
    steam: SteamClient
    job_futures: dict[str, asyncio.Future[Any]]

    def post_init(self):
        store_dir = self.xmpp.home_dir / self.user.bare_jid
        store_dir.mkdir(exist_ok=True)

        self.job_futures = {}

        self.steam = SteamClient()
        self.steam.set_credential_location(store_dir)
        self.steam.username = self.user.registration_form["username"]

    @staticmethod
    def xmpp_msg_id_to_legacy_msg_id(legacy_msg_id: Any) -> int:
        return int(legacy_msg_id)

    async def login(self):
        username = self.user.registration_form["username"]
        password = self.user.registration_form["password"]

        # self.steam.on(SteamClient.EVENT_CHAT_MESSAGE, self.on_steam_msg)
        self.steam.on(EMsg.ClientPersonaState, self.on_persona_state)
        self.steam.on("FriendMessagesClient.IncomingMessage#1", self.on_friend_message)
        self.steam.on("FriendMessagesClient.MessageReaction#1", self.on_friend_reaction)
        self.steam.on(EMsg.ServiceMethodResponse, self.on_service_method_response)

        login_result = self.steam.relogin()

        if login_result != EResult.OK:
            login_result = self.steam.login(username, password)

            if login_result == EResult.AccountLogonDenied:  # 2FA
                code = await self.input("Enter the code you received by email")
                login_result = self.steam.login(
                    self.user.registration_form["username"],
                    self.user.registration_form["password"],
                    auth_code=code,
                )

        self.log.debug("Login result: %s", login_result)
        if login_result == EResult.OK:
            self.log.debug("Login success")
        else:
            raise RuntimeError("Could not connect to steam")

        for f in self.steam.friends:
            f: SteamUser
            self.log.debug("Friend: %s - %s - %s", f, f.name, f.steam_id.id)
            c = self.contacts.by_legacy_id(f.steam_id.id)
            c.name = f.name
            c.avatar = f.get_avatar_url()
            await c.add_to_roster()
            c.update_status(f.state)

        asyncio.create_task(self.idle())

    async def idle(self):
        while True:
            self.steam.idle()
            await asyncio.sleep(0.1)

    def on_service_method_response(self, msg):
        self.log.debug("New service method response : %s", msg)
        try:
            fut = self.job_futures.pop(f"job_{msg.header.jobid_target}")
        except KeyError:
            self.log.debug(
                "Ignoring: %s vs %s", msg.header.jobid_target, self.job_futures
            )
        else:
            fut.set_result(msg.body)

    def on_friend_message(self, msg):
        self.log.debug("New friend message : %s", msg)
        if (type_ := msg.body.chat_entry_type) == steam.enums.EChatEntryType.Typing:
            user = self.steam.get_user(msg.body.steamid_friend)
            self.contacts.by_legacy_id(user.steam_id.id).composing()
        elif type_ == steam.enums.EChatEntryType.ChatMsg:
            user = self.steam.get_user(msg.body.steamid_friend)
            self.contacts.by_legacy_id(user.steam_id.id).send_text(
                msg.body.message, legacy_msg_id=msg.body.rtime32_server_timestamp
            )

    def on_friend_reaction(self, msg):
        self.log.debug("New friend reaction : %s", msg)
        if msg.body.reactor == self.steam.steam_id:
            pass
        else:
            if msg.body.reaction_type == k_EMessageReactionType_Emoticon:
                if msg.body.is_add:
                    # FIXME: this replace the XMPP reaction with the latest steam reaction
                    # we would need to fetch the friend's reaction list, but I did not find
                    # how to do that via this library
                    emoji = emoji_translate.get(msg.body.reaction, "❓")
                    self.contacts.by_legacy_id(SteamID(msg.body.reactor).id).react(
                        msg.body.server_timestamp, emoji
                    )
                else:
                    # FIXME: instead of retracting a single reaction, this deletes all reactions from the contact
                    # again, we need to retrieve the list of reactions for this message
                    self.contacts.by_legacy_id(SteamID(msg.body.reactor).id).react(
                        msg.body.server_timestamp, ""
                    )

    def on_persona_state(self, msg: MsgProto):
        persona_state = msg.body
        self.log.debug("New state event: %s", persona_state)
        for f in persona_state.friends:
            if f.friendid == self.steam.steam_id:
                self.log.debug("This is me %s", self.steam.steam_id)
                return
            self.contacts.by_legacy_id(SteamID(f.friendid).id).update_status(
                f.persona_state
            )

    async def logout(self):
        pass

    async def send_text(self, t: str, c: Contact, *, reply_to_msg_id=None):
        if not t:
            return
        job_id = self.steam.send_um(
            "FriendMessages.SendMessage#1",
            {
                "steamid": SteamID(c.legacy_id),
                "chat_entry_type": steam.enums.EChatEntryType.ChatMsg,
                "message": t,
            },
        )
        f = self.job_futures[job_id] = self.xmpp.loop.create_future()
        return (await f).server_timestamp

    async def send_file(self, u: str, c: Contact, *, reply_to_msg_id=None):
        return await self.send_text(u, c)

    async def active(self, c: Contact):
        pass

    async def inactive(self, c: Contact):
        pass

    async def composing(self, c: Contact):
        self.steam.send_um(
            "FriendMessages.SendMessage#1",
            {
                "steamid": SteamID(c.legacy_id),
                "chat_entry_type": steam.enums.EChatEntryType.Typing,
            },
        )

    async def paused(self, c: Contact):
        pass

    async def displayed(self, legacy_msg_id: Any, c: Contact):
        pass

    async def correct(self, text: str, legacy_msg_id: Any, c: Contact):
        pass

    async def search(self, form_values: dict[str, str]):
        pass

    async def react(self, legacy_msg_id: Any, emojis: list[str], c: Contact):
        for emoji in emojis:
            emoji_name = emoji_translate.inverse.get(emoji)
            if emoji_name is None:
                self.send_gateway_message(
                    f"On steam, you can only react with {' '.join(emoji_translate.values())}"
                )
                continue
            self.steam.send_um(
                "FriendMessages.UpdateMessageReaction#1",
                {
                    "steamid": SteamID(c.legacy_id).as_64,
                    "server_timestamp": legacy_msg_id,
                    "reaction_type": k_EMessageReactionType_Emoticon,
                    "reaction": emoji_name,
                    "is_add": True,
                },
            )

    async def retract(self, legacy_msg_id: Any, c: Contact):
        pass


emoji_translate = BiDict[str, str](
    [
        (":steamthumbsup:", "👍"),
        (":steamthumbsdown:", "👎"),
        (":steambored:", "🥱"),
        (":steamfacepalm:", "🤦"),
        (":steamhappy:", "😄"),
        (":steammocking:", "😝"),
        (":steamsalty:", "🧂"),
        (":steamsad:", "😔"),
        (":steamthis:", "⬆️"),
    ]
)



# isort: skip_file
# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin

from .stanza import Capabilities
from .static import StaticCaps
from .caps import XEP_0115

register_plugin(XEP_0115)



import asyncio
import logging
import re
import tempfile
from mimetypes import guess_type
from pathlib import Path
from typing import Optional

import aiohttp
import aiotdlib.api as tgapi
from aiotdlib.api.errors import BadRequest
from slixmpp.exceptions import XMPPError

from slidge import *

from .client import TelegramClient
from .contact import Contact, Roster
from .gateway import Gateway


class Session(BaseSession[Contact, Roster, Gateway]):
    tdlib_path: Optional[Path] = None
    tg: TelegramClient
    sent_read_marks: set[int]
    ack_futures: dict[int, asyncio.Future]
    user_correction_futures: dict[int, asyncio.Future]
    delete_futures: dict[int, asyncio.Future]

    def post_init(self):
        registration_form = {
            k: v if v != "" else None for k, v in self.user.registration_form.items()
        }
        self.sent_read_marks = set()
        self.ack_futures = {}
        self.user_correction_futures = {}
        self.delete_futures = {}

        i = registration_form.get("api_id")
        if i is not None:
            i = int(i)  # makes testing easier to make api_id optional...

        self.tg = TelegramClient(
            self,
            api_id=i,
            api_hash=registration_form.get("api_hash"),
            phone_number=registration_form["phone"],
            bot_token=registration_form.get("bot_token"),
            first_name=registration_form.get("first"),
            last_name=registration_form.get("last"),
            database_encryption_key=Gateway.args.tdlib_key,
            files_directory=Gateway.args.tdlib_path,
        )

    @staticmethod
    def xmpp_msg_id_to_legacy_msg_id(i: str) -> int:
        try:
            return int(i)
        except ValueError:
            raise NotImplementedError("This is not a valid telegram msg ID")

    async def login(self):
        await self.tg.start()
        await self.add_contacts_to_roster()
        return f"Connected as {await self.tg.get_my_id()}"

    async def logout(self):
        await self.tg.stop()

    async def wait_for_tdlib_success(self, result_id: int):
        fut = self.xmpp.loop.create_future()
        self.ack_futures[result_id] = fut
        return await fut

    async def send_text(self, t: str, c: "Contact", *, reply_to_msg_id=None) -> int:
        t = escape(t)
        try:
            result = await self.tg.send_text(
                chat_id=c.legacy_id, text=t, reply_to_message_id=reply_to_msg_id
            )
        except tgapi.BadRequest as e:
            if e.code == 400:
                raise XMPPError(condition="item-not-found", text="No such contact")
            else:
                raise
        new_message_id = await self.wait_for_tdlib_success(result.id)
        self.log.debug("Result: %s / %s", result, new_message_id)
        return new_message_id

    async def send_file(self, u: str, c: "Contact", *, reply_to_msg_id=None) -> int:
        type_, _ = guess_type(u)
        if type_ is not None:
            type_, subtype = type_.split("/")

        async with aiohttp.ClientSession() as session:
            async with session.get(u) as response:
                response.raise_for_status()
                with tempfile.NamedTemporaryFile() as file:
                    bytes_ = await response.read()
                    file.write(bytes_)
                    if type_ == "image":
                        result = await self.tg.send_photo(
                            chat_id=c.legacy_id, photo=file.name
                        )
                    elif type_ == "video":
                        result = await self.tg.send_video(
                            chat_id=c.legacy_id, video=file.name
                        )
                    else:
                        result = await self.tg.send_document(
                            c.legacy_id, document=file.name
                        )

        return result.id

    async def active(self, c: "Contact"):
        action = tgapi.OpenChat.construct(chat_id=c.legacy_id)
        res = await self.tg.request(action)
        self.log.debug("Open chat res: %s", res)

    async def inactive(self, c: "Contact"):
        action = tgapi.CloseChat.construct(chat_id=c.legacy_id)
        res = await self.tg.request(action)
        self.log.debug("Close chat res: %s", res)

    async def composing(self, c: "Contact"):
        action = tgapi.SendChatAction.construct(
            chat_id=c.legacy_id,
            action=tgapi.ChatActionTyping(),
            message_thread_id=0,  # TODO: check what telegram's threads really are
        )

        res = await self.tg.request(action)
        self.log.debug("Send composing res: %s", res)

    async def paused(self, c: "Contact"):
        pass

    async def displayed(self, tg_id: int, c: "Contact"):
        query = tgapi.ViewMessages.construct(
            chat_id=c.legacy_id,
            message_thread_id=0,
            message_ids=[tg_id],
            force_read=True,
        )
        res = await self.tg.request(query)
        self.log.debug("Send chat action res: %s", res)

    async def add_contacts_to_roster(self):
        chats = await self.tg.get_main_list_chats_all()
        for chat in chats:
            if not isinstance(chat.type_, tgapi.ChatTypePrivate):
                self.log.debug(
                    "Skipping %s as it is of type %s", chat.title, chat.type_
                )
            if isinstance(chat.photo, tgapi.ChatPhotoInfo):
                query = tgapi.DownloadFile.construct(
                    file_id=chat.photo.big.id, synchronous=True, priority=32
                )
                response: tgapi.File = await self.tg.request(query)
                with open(response.local.path, "rb") as f:
                    avatar = f.read()
            else:
                avatar = None
            contact = self.contacts.by_legacy_id(chat.id)
            contact.name = chat.title
            contact.avatar = avatar
            await contact.add_to_roster()
            contact.online()
            contact.away()

    async def correct(self, text: str, legacy_msg_id: int, c: "Contact"):
        f = self.user_correction_futures[legacy_msg_id] = self.xmpp.loop.create_future()
        query = tgapi.EditMessageText.construct(
            chat_id=c.legacy_id,
            message_id=legacy_msg_id,
            input_message_content=tgapi.InputMessageText.construct(
                text=tgapi.FormattedText.construct(text=text)
            ),
        )
        await self.tg.request(query)
        await f

    async def search(self, form_values: dict[str, str]):
        phone = form_values["phone"]
        response: tgapi.ImportedContacts = await self.tg.request(
            query=tgapi.ImportContacts(
                contacts=[
                    tgapi.Contact(
                        phone_number=phone,
                        user_id=0,
                        first_name=phone,
                        vcard="",
                        last_name="",
                    )
                ]
            )
        )
        user_id = response.user_ids[0]
        if user_id == 0:
            return

        await self.add_contacts_to_roster()
        contact = self.contacts.by_legacy_id(user_id)
        await contact.add_to_roster()

        return SearchResult(
            fields=[FormField("phone"), FormField("jid", type="jid-single")],
            items=[{"phone": form_values["phone"], "jid": contact.jid.bare}],
        )

    async def react(self, legacy_msg_id, emojis, c: "Contact"):
        if len(emojis) == 0:
            r = await self.tg.request(
                query=tgapi.SetMessageReaction(
                    chat_id=c.legacy_id,
                    message_id=legacy_msg_id,
                    reaction="",
                    is_big=False,
                )
            )
            self.log.debug("Remove reaction response: %s", r)
            return

        if len(emojis) > 1:
            c.carbon_react(legacy_msg_id)
            self.send_gateway_message(
                "Warning: unlike XMPP, telegram only accepts one reaction per message. "
                "Only your last reaction will be visible on Telegram."
            )

        emoji = emojis[-1]

        try:
            r = await self.tg.request(
                query=tgapi.SetMessageReaction(
                    chat_id=c.legacy_id,
                    message_id=legacy_msg_id,
                    reaction=emoji,
                    is_big=False,
                )
            )
        except BadRequest:
            available: tgapi.AvailableReactions = await self.tg.request(
                tgapi.GetMessageAvailableReactions(
                    chat_id=c.legacy_id, message_id=legacy_msg_id
                )
            )
            available_emojis = [a.reaction for a in available.reactions]
            self.send_gateway_message(
                "Error: unlike XMPP, telegram does not allow arbitrary emojis to be used as reactions. "
                f"Please pick your reaction in this list: {' '.join(available_emojis)}"
            )
            c.carbon_react(legacy_msg_id)
            raise XMPPError(
                "not-acceptable", text="This emoji is not allowed on telegram"
            )  # ignored by movim, unfortunately
        else:
            self.log.debug("Message reaction response: %s", r)

    async def retract(self, legacy_msg_id, c):
        f = self.delete_futures[legacy_msg_id] = self.xmpp.loop.create_future()
        r = await self.tg.request(
            tgapi.DeleteMessages(
                chat_id=c.legacy_id, message_ids=[legacy_msg_id], revoke=True
            )
        )
        self.log.debug("Delete message response: %s", r)
        confirmation = await f
        self.log.debug("Message delete confirmation: %s", confirmation)

    async def list_sessions(self) -> tgapi.Sessions:
        return await self.tg.request(tgapi.GetActiveSessions())

    async def terminate_session(self, session_id):
        return await self.tg.request(tgapi.TerminateSession(session_id=session_id))


def escape(t: str):
    return re.sub(ESCAPE_PATTERN, r"\\\1", t)


RESERVED_CHARS = "_*[]()~`>#+-=|{}.!"
ESCAPE_PATTERN = re.compile(f"([{re.escape(RESERVED_CHARS)}])")

log = logging.getLogger(__name__)



import asyncio
import logging
import os
import tempfile
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, Any

import aiohttp
import aiosignald.exc as sigexc
import aiosignald.generated as sigapi
from slixmpp.exceptions import XMPPError

from slidge import *

if TYPE_CHECKING:
    from .contact import Contact, Roster
    from .gateway import Gateway

from . import txt


class Session(BaseSession["Contact", "Roster", "Gateway"]):
    """
    Represents a signal account
    """

    # contacts: Roster

    def __init__(self, user: GatewayUser):
        """

        :param user:
        """
        super().__init__(user)
        self.phone = self.user.registration_form["phone"]
        if self.phone is None:
            raise RuntimeError
        self.signal = self.xmpp.signal
        self.xmpp.sessions_by_phone[self.phone] = self
        self.reaction_ack_futures: dict[tuple[int, str], asyncio.Future[None]] = {}

    @staticmethod
    def xmpp_msg_id_to_legacy_msg_id(i: str) -> int:
        try:
            return int(i)
        except ValueError:
            raise NotImplementedError

    async def paused(self, c: "Contact"):
        await (await self.signal).typing(
            account=self.phone, typing=False, address=c.signal_address
        )

    async def correct(self, text: str, legacy_msg_id: Any, c: "Contact"):
        return await self.send_text("Correction: " + text, c)

    async def search(self, form_values: dict[str, str]):
        pass

    async def login(self):
        """
        Attempt to listen to incoming events for this account,
        or pursue the registration process if needed.
        """
        try:
            await (await self.signal).subscribe(account=self.phone)
        except sigexc.NoSuchAccountError:
            device = self.user.registration_form["device"]
            try:
                if device == "primary":
                    await self.register()
                elif device == "secondary":
                    await self.link()
                else:
                    # This should never happen
                    self.send_gateway_status("Disconnected", show="dnd")
                    raise TypeError("Unknown device type", device)
            except sigexc.SignaldException as e:
                self.xmpp.send_message(
                    mto=self.user.jid,
                    mbody=f"Something went wrong: {e}",
                    mfrom=self.xmpp.boundjid,
                )
                raise
            await (await self.signal).subscribe(account=self.phone)

        return f"Connected as {self.phone}"

    async def register(self):
        self.send_gateway_status("Registering…", show="dnd")
        try:
            await (await self.signal).register(self.phone)
        except sigexc.CaptchaRequiredError:
            self.send_gateway_status("Captcha required", show="dnd")
            captcha = await self.input(txt.CAPTCHA_REQUIRED)
            await (await self.signal).register(self.phone, captcha=captcha)
        sms_code = await self.input(
            f"Reply to this message with the code you have received by SMS at {self.phone}.",
        )
        await (await self.signal).verify(account=self.phone, code=sms_code)
        await (await self.signal).set_profile(
            account=self.phone, name=self.user.registration_form["name"]
        )
        self.send_gateway_message(txt.REGISTER_SUCCESS)

    async def send_linking_qrcode(self):
        self.send_gateway_status("QR scan needed", show="dnd")
        resp = await (await self.signal).generate_linking_uri()
        await self.send_qr(resp.uri)
        self.xmpp.send_message(
            mto=self.user.jid,
            mbody=f"Use this URI or QR code on another signal device to "
            f"finish linking your XMPP account\n{resp.uri}",
            mfrom=self.xmpp.boundjid,
        )
        return resp

    async def link(self):
        resp = await self.send_linking_qrcode()
        try:
            await (await self.signal).finish_link(
                device_name=self.user.registration_form["device_name"],
                session_id=resp.session_id,
            )
        except sigexc.ScanTimeoutError:
            while True:
                r = await self.input(txt.LINK_TIMEOUT)
                if r in ("cancel", "link"):
                    break
                else:
                    self.send_gateway_message("Please reply either 'link' or 'cancel'")
            if r == "cancel":
                raise
            elif r == "link":
                await self.link()  # TODO: set a max number of attempts
        except sigexc.SignaldException as e:
            self.xmpp.send_message(
                mto=self.user.jid,
                mbody=f"Something went wrong during the linking process: {e}.",
                mfrom=self.xmpp.boundjid,
            )
            raise
        else:
            self.send_gateway_message(txt.LINK_SUCCESS)

    async def logout(self):
        await (await self.signal).unsubscribe(account=self.phone)

    async def add_contacts_to_roster(self):
        """
        Populate a user's roster
        """
        profiles = await (await self.signal).list_contacts(account=self.phone)
        for profile in profiles.profiles:
            full_profile = await (await self.signal).get_profile(
                account=self.phone, address=profile.address
            )
            contact = await self.contacts.by_json_address(profile.address)
            contact.name = profile.name or profile.profile_name
            if contact.name is not None:
                contact.name = contact.name.replace("\u0000", "")
            if full_profile.avatar is not None:
                with open(full_profile.avatar, "rb") as f:
                    contact.avatar = f.read()
            await contact.add_to_roster()
            contact.online()

    async def on_signal_message(self, msg: sigapi.IncomingMessagev1):
        """
        User has received 'something' from signal

        :param msg:
        """
        if (sync_msg := msg.sync_message) is not None:
            if sync_msg.contacts is not None and msg.sync_message.contactsComplete:
                log.debug("Received a sync contact updates")
                await self.add_contacts_to_roster()

            if (sent := sync_msg.sent) is None:
                # Probably a 'message read' marker
                log.debug("No sent message in this sync message")
                return
            contact = await self.contacts.by_json_address(sent.destination)
            sent_msg = sent.message

            if (body := sent_msg.body) is not None:
                contact.carbon(
                    body=body,
                    date=datetime.fromtimestamp(sent_msg.timestamp / 1000),
                )
            if (reaction := sent_msg.reaction) is not None:
                try:
                    fut = self.reaction_ack_futures.pop(
                        (reaction.targetSentTimestamp, reaction.emoji)
                    )
                except KeyError:
                    contact.carbon_react(
                        reaction.targetSentTimestamp,
                        () if reaction.remove else reaction.emoji,
                    )
                else:
                    fut.set_result(None)
            if (delete := sent_msg.remoteDelete) is not None:
                contact.carbon_retract(delete.target_sent_timestamp)

        contact = await self.contacts.by_json_address(msg.source)

        if (data := msg.data_message) is not None:
            if (quote := data.quote) is None:
                reply_to_msg_id = None
            else:
                reply_to_msg_id = quote.id
            for attachment in data.attachments:
                with open(attachment.storedFilename, "rb") as f:
                    await contact.send_file(
                        filename=attachment.customFilename,
                        input_file=f,
                        content_type=attachment.contentType,
                        legacy_msg_id=msg.data_message.timestamp,
                        reply_to_msg_id=reply_to_msg_id,
                    )
            if (body := data.body) is not None:
                contact.send_text(
                    body=body,
                    legacy_msg_id=msg.data_message.timestamp,
                    reply_to_msg_id=reply_to_msg_id,
                )
            if (reaction := data.reaction) is not None:
                self.log.debug("Reaction: %s", reaction)
                if reaction.remove:
                    contact.react(reaction.targetSentTimestamp)
                else:
                    contact.react(reaction.targetSentTimestamp, reaction.emoji)
            if (delete := data.remoteDelete) is not None:
                contact.retract(delete.target_sent_timestamp)

        if (typing_message := msg.typing_message) is not None:
            action = typing_message.action
            if action == "STARTED":
                contact.active()
                contact.composing()
            elif action == "STOPPED":
                contact.paused()

        if (receipt_message := msg.receipt_message) is not None:
            type_ = receipt_message.type
            if type_ == "DELIVERY":
                for t in msg.receipt_message.timestamps:
                    contact.received(t)
            elif type_ == "READ":
                for t in msg.receipt_message.timestamps:
                    contact.displayed(t)

    async def send_text(self, t: str, c: "Contact", *, reply_to_msg_id=None) -> int:
        if reply_to_msg_id is None:
            quote = None
        else:
            quote = sigapi.JsonQuotev1(
                id=reply_to_msg_id,
                author=c.signal_address,
                text=""  # not sure what this accomplishes? does not seem to have any effect,
                # but must not be None or NullPointerException
            )
        response = await (await self.signal).send(
            account=self.phone,
            recipientAddress=c.signal_address,
            messageBody=t,
            quote=quote,
        )
        result = response.results[0]
        log.debug("Result: %s", result)
        if result.networkFailure or result.proof_required_failure:
            raise XMPPError(str(result))
        elif result.identityFailure:
            s = await self.signal
            identities = (
                await s.get_identities(
                    account=self.phone,
                    address=c.signal_address,
                )
            ).identities
            ans = await self.input(
                f"The identity of {c.phone} has changed. "
                f"Do you want to trust all their identities and resend the message?"
            )
            if ans.lower().startswith("y"):
                for i in identities:
                    await (await self.signal).trust(
                        account=self.phone,
                        address=c.signal_address,
                        safety_number=i.safety_number,
                    )
                await self.send_text(t, c, reply_to_msg_id=reply_to_msg_id)
            else:
                raise XMPPError(str(result))
        return response.timestamp

    async def send_file(self, u: str, c: "Contact", *, reply_to_msg_id=None):
        s = await self.signal
        async with aiohttp.ClientSession() as client:
            async with client.get(url=u) as r:
                with tempfile.TemporaryDirectory(
                    dir=Path(self.xmpp.signal_socket).parent,
                ) as d:
                    os.chmod(d, 0o777)
                    with open(Path(d) / r.url.name, "wb") as f:
                        f.write(await r.content.read())
                        os.chmod(
                            f.name, 0o666
                        )  # temp file is 0600 https://stackoverflow.com/a/10541972/5902284
                        signal_r = await s.send(
                            account=self.phone,
                            recipientAddress=c.signal_address,
                            attachments=[sigapi.JsonAttachmentv1(filename=f.name)],
                        )
                        return signal_r.timestamp

    async def active(self, c: "Contact"):
        pass

    async def inactive(self, c: "Contact"):
        pass

    async def composing(self, c: "Contact"):
        await (await self.signal).typing(
            account=self.phone,
            address=c.signal_address,
            typing=True,
        )

    async def displayed(self, legacy_msg_id: int, c: "Contact"):
        await (await self.signal).mark_read(
            account=self.phone,
            to=c.signal_address,
            timestamps=[legacy_msg_id],
        )

    async def react(self, legacy_msg_id: int, emojis: list[str], c: "Contact"):
        remove = len(emojis) == 0
        if len(emojis) == 0:
            remove = True
            emoji = ""
        else:
            emoji = emojis[-1]
            if len(emojis) > 1:
                self.send_gateway_message("Only one reaction per message on signal")
                c.carbon_react(legacy_msg_id, emoji)

        response = await (await self.signal).react(
            username=self.phone,
            recipientAddress=c.signal_address,
            reaction=sigapi.JsonReactionv1(
                emoji=emoji,
                remove=remove,
                targetAuthor=sigapi.JsonAddressv1(number=self.phone)
                if legacy_msg_id in self.sent
                else c.signal_address,
                targetSentTimestamp=legacy_msg_id,
            ),
        )
        result = response.results[0]
        if (
            result.networkFailure
            or result.identityFailure
            or result.proof_required_failure
        ):
            raise XMPPError(str(result))
        f = self.reaction_ack_futures[
            (legacy_msg_id, emoji)
        ] = self.xmpp.loop.create_future()
        await f

    async def retract(self, legacy_msg_id: int, c: "Contact"):
        try:
            await (await self.signal).remote_delete(
                account=self.phone, address=c.signal_address, timestamp=legacy_msg_id
            )
        except sigexc.SignaldException as e:
            raise XMPPError(text=f"Something went wrong during remote delete: {e}")

    async def add_device(self, uri: str):
        try:
            await (await self.signal).add_device(account=self.phone, uri=uri)
        except sigexc.SignaldException as e:
            self.send_gateway_message(f"Problem: {e}")
        else:
            self.send_gateway_message("Linking OK")


log = logging.getLogger(__name__)



import functools
from typing import Optional

import discord as di

from slidge import *

from .session import Session


class Gateway(BaseGateway[Session]):
    COMPONENT_NAME = "Discord (slidge)"
    REGISTRATION_INSTRUCTIONS = (
        "Have a look at https://discordpy-self.readthedocs.io/en/latest/token.html"
    )
    REGISTRATION_FIELDS = [FormField("token", required=True)]

    ROSTER_GROUP = "Discord"


class Contact(LegacyContact[Session]):
    MARKS = False

    def __init__(self, *a, **k):
        super().__init__(*a, **k)
        self._discord_id: Optional[int] = None

    @property
    def discord_id(self):
        if self._discord_id is None:
            for u in self.session.discord.users:
                if not isinstance(u, di.User):
                    continue
                if str(u) == self.legacy_id:
                    self._discord_id = u.id
                    break
        return self._discord_id

    @discord_id.setter
    def discord_id(self, i: int):
        self._discord_id = i

    @functools.cached_property
    def discord_user(self) -> di.User:
        return self.session.discord.get_user(self._discord_id)

    @functools.cached_property
    def direct_channel_id(self):
        return self.discord_user.dm_channel.id

    async def update_reactions(self, m: di.Message):
        legacy_reactions = []
        user = self.discord_user
        for r in m.reactions:
            async for u in r.users():
                if u == user:
                    legacy_reactions.append(r.emoji)
        self.react(m.id, legacy_reactions)


class Roster(LegacyRoster[Contact, "Session"]):
    def by_discord_user(self, u: di.User):
        return self.by_legacy_id(str(u))



from .util import (
    ABCSubclassableOnceAtMost,
    BiDict,
    FormField,
    SearchResult,
    SubclassableOnce,
)

__all__ = [
    "BiDict",
    "SearchResult",
    "FormField",
    "SubclassableOnce",
    "ABCSubclassableOnceAtMost",
]



import dataclasses
import logging
from abc import ABCMeta
from typing import Generic, Iterable, Literal, Optional, TypeVar

field_type = Literal[
    "boolean",
    "fixed",
    "text-single",
    "jid-single",
    "list-single",
    "list-multi",
    "text-private",
]


@dataclasses.dataclass
class FormField:
    """
    Represents a field of the form that a user will see when registering to the gateway
    via their XMPP client.
    """

    var: str = ""
    """
    Internal name of the field, will be used to retrieve via :py:attr:`slidge.GatewayUser.registration_form`
    """
    label: Optional[str] = None
    """Description of the field that the aspiring user will see"""
    required: bool = False
    """Whether this field is mandatory or not"""
    private: bool = False
    """
    For sensitive info that should not be displayed on screen while the user types.
    Forces field_type to "text-private"
    """
    type: field_type = "text-single"
    """Type of the field, see `XEP-0004 <https://xmpp.org/extensions/xep-0004.html#protocol-fieldtypes>`_"""
    value: str = ""
    """Pre-filled value. Will be automatically pre-filled if a registered user modifies their subscription"""
    options: Optional[list[dict[str, str]]] = None

    def dict(self):
        return dataclasses.asdict(self)

    def __post_init__(self):
        if self.private:
            self.type = "text-private"


KeyType = TypeVar("KeyType")  # FIXME: mypy does not correctly infer types with this...
ValueType = TypeVar("ValueType")


class BiDict(Generic[KeyType, ValueType], dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.inverse: dict[ValueType, KeyType] = {}
        for key, value in self.items():
            self.inverse[value] = key

    def __setitem__(self, key: KeyType, value: ValueType):
        if key in self:
            del self.inverse[self[key]]
        super().__setitem__(key, value)
        self.inverse[value] = key


@dataclasses.dataclass
class SearchResult:
    fields: Iterable[FormField]
    items: Iterable[dict[str, str]]


class SubclassableOnce(type):
    TEST_MODE = False  # To allow importing everything, including plugins, during tests

    def __init__(cls, name, bases, dct):
        for b in bases:
            if type(b) in (SubclassableOnce, ABCSubclassableOnceAtMost):
                if hasattr(b, "_subclass") and not cls.TEST_MODE:
                    raise RuntimeError(
                        "This class must be subclassed once at most!",
                        cls,
                        name,
                        bases,
                        dct,
                    )
                else:
                    log.debug("Setting %s as subclass for %s", cls, b)
                    b._subclass = cls

        super().__init__(name, bases, dct)

    def get_self_or_unique_subclass(cls):
        try:
            return cls.get_unique_subclass()
        except AttributeError:
            return cls

    def get_unique_subclass(cls):
        r = getattr(cls, "_subclass", None)
        if r is None:
            raise AttributeError("Could not find any subclass", cls)
        return r

    def reset_subclass(cls):
        try:
            log.debug("Resetting subclass of %s", cls)
            delattr(cls, "_subclass")
        except AttributeError:
            log.debug("No subclass were registered for %s", cls)


class ABCSubclassableOnceAtMost(ABCMeta, SubclassableOnce):
    pass


log = logging.getLogger(__name__)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['slidge',
 'slidge.core',
 'slidge.plugins',
 'slidge.plugins.discord_self',
 'slidge.plugins.mattermost',
 'slidge.plugins.signal',
 'slidge.plugins.telegram',
 'slidge.util',
 'slidge.util.xep_0030',
 'slidge.util.xep_0030.stanza',
 'slidge.util.xep_0055',
 'slidge.util.xep_0077',
 'slidge.util.xep_0100',
 'slidge.util.xep_0115',
 'slidge.util.xep_0333',
 'slidge.util.xep_0356',
 'slidge.util.xep_0356_old',
 'slidge.util.xep_0363',
 'slidge.util.xep_0461']

package_data = \
{'': ['*']}

install_requires = \
['ConfigArgParse>=1.5.3,<2.0.0',
 'Pillow>=8.1.0',
 'aiohttp>=3.6.0',
 'pickle-secure>=0.9.99,<0.10.0',
 'qrcode>=7.3',
 'slixmpp>=1.8.2,<2.0.0']

extras_require = \
{'discord': ['discord.py-self>=1.9.2,<2.0.0'],
 'facebook': ['mautrix-facebook>=0.4.0,<0.5.0'],
 'mattermost': ['mattermost-api-reference-client>=4.0.0,<5.0.0',
                'emoji>=2.0.0,<3.0.0'],
 'signal': ['aiosignald>=0.3.1,<0.4.0'],
 'skype': ['SkPy>=0.10.4,<0.11.0'],
 'steam': ['steam[client]>=1.3.0,<2.0.0'],
 'telegram': ['aiotdlib>=0.19.2,<0.20.0', 'pydantic']}

entry_points = \
{'console_scripts': ['slidge = slidge.__main__:main']}

setup_kwargs = {
    'name': 'slidge',
    'version': '0.1.0a4',
    'description': 'XMPP bridging framework',
    'long_description': 'Slidge 🛷\n========\n\n[Home](https://sr.ht/~nicoco/slidge) |\n[Source](https://sr.ht/~nicoco/slidge/sources) |\n[Issues](https://sr.ht/~nicoco/slidge/trackers) |\n[Patches](https://lists.sr.ht/~nicoco/public-inbox) |\n[Chat](xmpp:slidge@conference.nicoco.fr?join)\n\nTurn any XMPP client into that fancy multiprotocol chat app that every cool kid want.\n\n[![Documentation status](https://readthedocs.org/projects/slidge/badge/?version=latest)](https://slidge.readthedocs.io/)\n[![builds.sr.ht status](https://builds.sr.ht/~nicoco/slidge/commits/master/.build.yml.svg)](https://builds.sr.ht/~nicoco/slidge/commits/master/.build.yml?)\n[![pypi](https://badge.fury.io/py/slidge.svg)](https://pypi.org/project/slidge/)\n\nSlidge is a general purpose XMPP (puppeteer) gateway framework in python.\nIt\'s a work in progress, but it should make\n[writing gateways to other chat networks](https://slidge.readthedocs.io/en/latest/dev/tutorial.html)\n(*plugins*) as frictionless as possible.\n\nIt comes with a few plugins included, implementing at least basic direct messaging and often more "advanced"\ninstant messaging features:\n\n|            | ⏻[¹] | …[²] | ✓[³] | 🗎[⁴] | ✎[⁵] | ☺[⁶] | 🗑[⁷]  | ↵[⁸] | \n|------------|------|------|------|-------|------|------|--------|------|\n| Signal     | -    | ✓    | ✓    | ✓     | -    | ✓    | ✓      | ✓    |\n| Telegram   | ✓    | ✓    | ✓    | ✓     | ✓    | ✓    | ✓      | ✓    |\n| Discord    | ✗    | ✓    | -    | ✓     | ✓    | ~    | ✓      | ✓    |\n| Steam      | ✓    | ✓    | -    | ✗     | -    | ~    | -      | -    |\n| Mattermost | ~    | ✓    | -    | ✓     | ✓    | ✓    | ✓      | ✗    |\n| Facebook   | ✗    | ✓    | ✓    | ✓     | ✓    | ✓    | ✓      | ✓    |\n| Skype      | ✗    | ✗    | ✗    | ~     | ✗    | ✗    | ✗      | ✗    |\n\n\n[¹]: https://xmpp.org/rfcs/rfc6121.html#presence\n[²]: https://xmpp.org/extensions/xep-0085.html\n[³]: https://xmpp.org/extensions/xep-0333.html\n[⁴]: https://xmpp.org/extensions/xep-0363.html\n[⁵]: https://xmpp.org/extensions/xep-0308.html\n[⁶]: https://xmpp.org/extensions/xep-0444.html\n[⁷]: https://xmpp.org/extensions/xep-0424.html\n[⁸]: https://xmpp.org/extensions/xep-0461.html\n\n\n(this table may not be entirely accurate, but **in theory**, stuff marked ✓ works)\n\nNB: - means that the legacy network does not have an equivalent of this XMPP feature\n    (because XMPP is better, what did you think?)\n\n**WARNING**: you may break the terms of use of a legacy network and end up getting your account locked\nby using slidge. Refer to the [keeping a low profile](https://slidge.readthedocs.io/en/latest/user/low_profile.html)\ndocumentation page for more info.\n\nStatus\n------\n\nSlidge is alpha-grade software.\nRight now, only direct messages are implemented, no group chat stuff at all.\nDirect messaging does (more or less) work though.\nAny contribution whatsoever (testing, patches, suggestions, beer, …) is more than welcome.\nDon\'t be shy!\n\nTesting locally should be fairly easy, so please go ahead and give me some\nfeedback, through the [MUC](xmpp:slidge@conference.nicoco.fr?join), the\n[issue tracker](https://todo.sr.ht/~nicoco/slidge) or in my\n[public inbox](https://lists.sr.ht/~nicoco/public-inbox).\n\nInstallation\n------------\n\n#### docker-compose\n\nDocker-compose spins up a local XMPP server preconfigured for you., with a ``test@localhost`` / ``password``\naccount\n\n```sh\ndocker-compose up\n```\n\nFor the other options, you need a\n[configured](https://slidge.readthedocs.io/en/latest/admin/general.html#configure-the-xmpp-server)\nXMPP server.\n\n#### poetry\n\n```sh\npoetry install --extras signal  # you can replace signal with any network listed in the table above\npoetry run python -m slidge --legacy-module=slidge.plugins.signal\n```\n\n#### pip\n\n```sh\npip install slidge[signal]  # you can replace signal with any network listed in the table above\npython -m slidge --legacy-module=slidge.plugins.signal\n```\n\n### XMPP client\n\n#### movim\n\nIf you used docker-compose, you should be able to use the [movim](https://movim.eu) client\nfrom your browser at http://localhost:8888\n\nUnfortunately, the movim UI thinks that ``test@localhost`` is not a valid JID and does not let you click\non the "Connect" button.\nAs a workaround, use your browser dev tools to inspect and modify the ``<input id="username"`` in order to\nremove the ``pattern="^[^...`` attribute.\n\nThen go to the Configuration/Account tab. You should be able to register to the slidge gateways from here.\n\n#### Gajim\n\nInstall and launch [gajim](https://gajim.org) and add your XMPP account.\nGo to "Accounts"→"Discover services".\nYou should see the slidge gateways as server components.\n\nAbout privacy\n-------------\n\nSlidge (and most if not all XMPP gateway that I know of) will break\nend-to-end encryption, or more precisely one of the \'ends\' become the\ngateway itself. If privacy is a major concern for you, my advice would\nbe to:\n\n-   use XMPP + OMEMO\n-   self-host your gateways\n-   have your gateways hosted by someone you know AFK and trust\n\nRelated projects\n----------------\n\n-   [Spectrum](https://www.spectrum.im/)\n-   [Bitfrost](https://github.com/matrix-org/matrix-bifrost)\n-   [Mautrix](https://github.com/mautrix)\n-   [matterbridge](https://github.com/42wim/matterbridge)\n-   [XMPP-discord-bridge](https://git.polynom.me/PapaTutuWawa/xmpp-discord-bridge)\n',
    'author': 'Nicolas Cedilnik',
    'author_email': 'nicoco@nicoco.fr',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://sr.ht/~nicoco/slidge/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import asyncio
import functools
from datetime import datetime
from typing import TYPE_CHECKING

import aiotdlib
from aiotdlib import api as tgapi

if TYPE_CHECKING:
    from .session import Session


class TelegramClient(aiotdlib.Client):
    def __init__(self, session: "Session", **kw):
        super().__init__(parse_mode=aiotdlib.ClientParseMode.MARKDOWN, **kw)
        self.session = session
        self.contacts = session.contacts
        self.log = self.session.log

        async def input_(prompt):
            self.session.send_gateway_status(f"Action required: {prompt}")
            return await session.input(prompt)

        self.input = input_
        self._auth_get_code = functools.partial(input_, "Enter code")
        self._auth_get_password = functools.partial(input_, "Enter 2FA password:")
        self._auth_get_first_name = functools.partial(input_, "Enter first name:")
        self._auth_get_last_name = functools.partial(input_, "Enter last name:")

        self.add_event_handler(self.dispatch_update, tgapi.API.Types.ANY)

    async def dispatch_update(self, _self, update: tgapi.Update):
        try:
            handler = getattr(self, "handle_" + update.ID[6:])
        except AttributeError:
            self.session.log.debug("No handler for %s, ignoring", update.ID)
        except IndexError:
            self.session.log.debug("Ignoring weird event: %s", update.ID)
        else:
            await handler(update)

    async def handle_NewMessage(self, update: tgapi.UpdateNewMessage):
        if (msg := update.message).is_channel_post:
            self.log.debug("Ignoring channel post")
            return

        if not await self.is_private_chat(msg.chat_id):
            self.log.debug("Ignoring group message")
            return

        session = self.session
        if msg.is_outgoing:
            # This means slidge is responsible for this message, so no carbon is needed;
            # but maybe this does not handle all possible cases gracefully?
            if msg.sending_state is not None or msg.id in session.sent:
                return
            content = msg.content
            if isinstance(content, tgapi.MessageText):
                session.contacts.by_legacy_id(msg.chat_id).carbon(
                    content.text.text, msg.id, datetime.fromtimestamp(msg.date)
                )
            # TODO: implement carbons for other contents
            return

        sender = msg.sender_id
        if not isinstance(sender, tgapi.MessageSenderUser):
            self.log.debug("Ignoring non-user sender")  # Does this happen?
            return

        await session.contacts.by_legacy_id(sender.user_id).send_tg_message(msg)

    async def handle_UserStatus(self, update: tgapi.UpdateUserStatus):
        if update.user_id == await self.get_my_id():
            return
        contact = self.contacts.by_legacy_id(update.user_id)
        if not contact.added_to_roster:
            self.log.debug("Ignoring presence of contact not in the roster")
            return
        await contact.send_tg_status(update.status)

    async def handle_ChatReadOutbox(self, update: tgapi.UpdateChatReadOutbox):
        if not await self.is_private_chat(update.chat_id):
            return
        self.contacts.by_legacy_id(update.chat_id).displayed(
            update.last_read_outbox_message_id
        )

    async def handle_ChatAction(self, action: tgapi.UpdateChatAction):
        if not await self.is_private_chat(action.chat_id):
            return

        sender = action.sender_id
        if not isinstance(sender, tgapi.MessageSenderUser):
            self.log.debug("Ignoring action: %s", action)
            return

        if (chat_id := action.chat_id) != sender.user_id:
            self.log.debug("Ignoring group (?) action: %s", action)
            return

        self.contacts.by_legacy_id(chat_id).composing()

    async def handle_ChatReadInbox(self, action: tgapi.UpdateChatReadInbox):
        if not await self.is_private_chat(action.chat_id):
            return

        session = self.session
        msg_id = action.last_read_inbox_message_id
        self.log.debug(
            "Self read mark for %s and we sent %s", msg_id, session.sent_read_marks
        )
        try:
            session.sent_read_marks.remove(msg_id)
        except KeyError:
            # slidge didn't send this read mark, so it comes from the official tg client
            contact = session.contacts.by_legacy_id(action.chat_id)
            contact.carbon_read(msg_id)

    async def handle_MessageContent(self, action: tgapi.UpdateMessageContent):
        if not await self.is_private_chat(action.chat_id):
            return

        new = action.new_content
        if not isinstance(new, tgapi.MessageText):
            raise NotImplementedError(new)
        session = self.session
        try:
            fut = session.user_correction_futures.pop(action.message_id)
        except KeyError:
            contact = session.contacts.by_legacy_id(action.chat_id)
            if action.message_id in self.session.sent:
                contact.carbon_correct(action.message_id, new.text.text)
            else:
                contact.correct(action.message_id, new.text.text)
        else:
            self.log.debug("User correction confirmation received")
            fut.set_result(None)

    async def handle_User(self, action: tgapi.UpdateUser):
        u = action.user
        if u.id == await self.get_my_id():
            return
        await self.request(
            query=tgapi.ImportContacts(
                contacts=[
                    tgapi.Contact(
                        phone_number=u.phone_number,
                        user_id=u.id,
                        first_name=u.first_name,
                        last_name=u.last_name,
                        vcard="",
                    )
                ]
            )
        )
        contact = self.session.contacts.by_legacy_id(u.id)
        contact.name = u.first_name
        await contact.add_to_roster()

    async def handle_MessageInteractionInfo(
        self, update: tgapi.UpdateMessageInteractionInfo
    ):
        if not await self.is_private_chat(update.chat_id):
            return

        contact = self.session.contacts.by_legacy_id(update.chat_id)
        me = await self.get_my_id()
        if update.interaction_info is None:
            contact.react(update.message_id, [])
        else:
            for reaction in update.interaction_info.reactions:
                for sender in reaction.recent_sender_ids:
                    if isinstance(sender, tgapi.MessageSenderUser):
                        if sender.user_id == contact.legacy_id:
                            contact.react(update.message_id, [reaction.reaction])
                        elif sender.user_id == me:
                            contact.carbon_react(update.message_id, [reaction.reaction])

    async def handle_DeleteMessages(self, update: tgapi.UpdateDeleteMessages):
        if not await self.is_private_chat(update.chat_id):
            return

        if not update.is_permanent:  # tdlib send 'delete from cache' updates apparently
            self.log.debug("Ignoring non permanent delete")
            return
        for legacy_msg_id in update.message_ids:
            try:
                future = self.session.delete_futures.pop(legacy_msg_id)
            except KeyError:
                # FIXME: where do we filter out group chat messages here ?!
                contact = self.session.contacts.by_legacy_id(update.chat_id)
                if legacy_msg_id in self.session.sent:
                    contact.carbon_retract(legacy_msg_id)
                else:
                    contact.retract(legacy_msg_id)
            else:
                future.set_result(update)

    async def handle_MessageSendSucceeded(
        self, update: tgapi.UpdateMessageSendSucceeded
    ):
        self.session.sent_read_marks.add(update.message.id)
        for _ in range(10):
            try:
                future = self.session.ack_futures.pop(update.message.id)
            except KeyError:
                await asyncio.sleep(0.5)
            else:
                future.set_result(update.message.id)
                return
        self.log.warning("Ignoring Send success for %s", update.message.id)

    async def is_private_chat(self, chat_id: int):
        chat = await self.get_chat(chat_id)
        return isinstance(chat, tgapi.ChatTypePrivate)



# slixmpp: The Slick XMPP Library
# Copyright (C) 2016 Emmanuel Gil Peyrot
# This file is part of slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin

from .markers import XEP_0333
from .stanza import Acknowledged, Displayed, Markable, Received

register_plugin(XEP_0333)



import asyncio
import concurrent.futures
import io
import logging
import pprint
from pathlib import Path
from threading import Lock, Thread
from typing import Any, Optional

import aiohttp
import skpy
from slixmpp import JID

from slidge import *


class Gateway(BaseGateway):
    REGISTRATION_INSTRUCTIONS = "Enter skype credentials"
    REGISTRATION_FIELDS = [
        FormField(var="username", label="Username", required=True),
        FormField(var="password", label="Password", required=True, private=True),
    ]

    ROSTER_GROUP = "Skype"

    COMPONENT_NAME = "Skype (slidge)"
    COMPONENT_TYPE = "skype"

    COMPONENT_AVATAR = "https://logodownload.org/wp-content/uploads/2017/05/skype-logo-1-1-2048x2048.png"

    async def validate(
        self, user_jid: JID, registration_form: dict[str, Optional[str]]
    ):
        pass


class Roster(LegacyRoster):
    # ':' is forbidden in the username part of a JID

    @staticmethod
    def legacy_id_to_jid_username(legacy_id: str) -> str:
        if legacy_id.startswith("live:"):
            return legacy_id.replace("live:", "__live__")
        else:
            return legacy_id

    @staticmethod
    def jid_username_to_legacy_id(jid_username: str) -> str:
        if jid_username.startswith("__live__"):
            return jid_username.replace("__live__", "live:")
        else:
            return jid_username


class Contact(LegacyContact):
    pass


class Session(BaseSession[Contact, Roster, Gateway]):
    skype_token_path: Path
    sk: skpy.Skype
    thread: Optional[Thread]
    sent_by_user_to_ack: dict[int, asyncio.Future]
    unread_by_user: dict[int, skpy.SkypeMsg]
    send_lock: Lock

    def post_init(self):
        self.skype_token_path = self.xmpp.home_dir / self.user.bare_jid
        self.thread = None
        self.sent_by_user_to_ack = {}
        self.unread_by_user = {}
        self.send_lock = Lock()

    async def async_wrap(self, func, *args):
        return await self.xmpp.loop.run_in_executor(executor, func, *args)

    async def login(self):
        f = self.user.registration_form
        try:
            self.sk = await self.async_wrap(
                skpy.Skype,
                f["username"],
                f["password"],
                str(self.skype_token_path),
            )
        except skpy.core.SkypeApiException:
            # workaround for https://github.com/Terrance/SkPy/issues/164
            # not sure why, but I need this for my (nicoco's) account
            # FWIW, I have a live (I think) a account with a very old skype account (pre-microsoft)
            # and I set up 2FA + app password for slidge
            self.sk = await self.async_wrap(skpy.Skype)
            self.sk.conn.setTokenFile(str(self.skype_token_path))
            self.sk.conn.soapLogin(f["username"], f["password"])

        # self.sk.subscribePresence()
        for contact in self.sk.contacts:
            c = self.contacts.by_legacy_id(contact.id)
            first = contact.name.first
            last = contact.name.last
            if first is not None and last is not None:
                c.name = f"{first} {last}"
            elif first is not None:
                c.name = first
            elif last is not None:
                c.name = last
            if contact.avatar is not None:
                c.avatar = contact.avatar
            await c.add_to_roster()
            c.online()
        # TODO: close this gracefully on exit
        self.thread = thread = Thread(target=self.skype_blocking)
        thread.start()
        return f"Connected as '{self.sk.userId}'"

    def skype_blocking(self):
        while True:
            for event in self.sk.getEvents():
                # no need to sleep since getEvents blocks for 30 seconds already
                asyncio.run_coroutine_threadsafe(
                    self.on_skype_event(event), self.xmpp.loop
                )

    async def on_skype_event(self, event: skpy.SkypeEvent):
        log.debug("Skype event: %s", event)
        if isinstance(event, skpy.SkypeNewMessageEvent):
            while self.send_lock.locked():
                await asyncio.sleep(0.1)
            msg = event.msg
            chat = event.msg.chat
            if isinstance(chat, skpy.SkypeSingleChat):
                log.debug("this is a single chat with user: %s", chat.userIds[0])
                contact = self.contacts.by_legacy_id(chat.userIds[0])
                if msg.userId == self.sk.userId:
                    try:
                        fut = self.sent_by_user_to_ack.pop(msg.id)
                    except KeyError:
                        if log.isEnabledFor(logging.DEBUG):
                            log.debug(
                                "Slidge did not send this message: %s",
                                pprint.pformat(vars(event)),
                            )
                        contact.carbon(msg.plain)
                    else:
                        fut.set_result(msg)
                else:
                    if isinstance(msg, skpy.SkypeTextMsg):
                        contact.send_text(msg.plain, legacy_msg_id=msg.id)
                        self.unread_by_user[msg.id] = msg
                    elif isinstance(msg, skpy.SkypeFileMsg):
                        file = io.BytesIO(
                            await self.async_wrap(lambda: msg.fileContent)
                        )  # non-blocking download / lambda because fileContent = property
                        await contact.send_file(filename=msg.file.name, input_file=file)
        elif isinstance(event, skpy.SkypeTypingEvent):
            contact = self.contacts.by_legacy_id(event.userId)
            if event.active:
                contact.composing()
            else:
                contact.paused()
        elif isinstance(event, skpy.SkypeChatUpdateEvent):
            if log.isEnabledFor(logging.DEBUG):
                log.debug("chat update: %s", pprint.pformat(vars(event)))
        # No 'contact has read' event :( https://github.com/Terrance/SkPy/issues/206
        await self.async_wrap(event.ack)

    async def send_text(self, t: str, c: LegacyContact, *, reply_to_msg_id=None):
        chat = self.sk.contacts[c.legacy_id].chat
        self.send_lock.acquire()
        msg = await self.async_wrap(chat.sendMsg, t)
        if log.isEnabledFor(logging.DEBUG):
            log.debug("Sent msg: %s", pprint.pformat(vars(msg)))
        future = asyncio.Future[skpy.SkypeMsg]()
        self.sent_by_user_to_ack[msg.id] = future
        self.send_lock.release()
        skype_msg = await future
        return skype_msg.id

    async def logout(self):
        pass

    async def send_file(self, u: str, c: LegacyContact, *, reply_to_msg_id=None):
        async with aiohttp.ClientSession() as session:
            async with session.get(u) as response:
                file_bytes = await response.read()
        fname = u.split("/")[-1]
        fname_lower = fname.lower()
        await self.async_wrap(
            self.sk.contacts[c.legacy_id].chat.sendFile,
            io.BytesIO(file_bytes),
            fname,
            any(fname_lower.endswith(x) for x in (".png", ".jpg", ".gif", ".jpeg")),
        )

    async def active(self, c: LegacyContact):
        pass

    async def inactive(self, c: LegacyContact):
        pass

    async def composing(self, c: LegacyContact):
        executor.submit(self.sk.contacts[c.legacy_id].chat.setTyping, True)

    async def paused(self, c: LegacyContact):
        executor.submit(self.sk.contacts[c.legacy_id].chat.setTyping, False)

    async def displayed(self, legacy_msg_id: int, c: LegacyContact):
        try:
            skype_msg = self.unread_by_user.pop(legacy_msg_id)
        except KeyError:
            log.debug(
                "We did not transmit: %s (%s)", legacy_msg_id, self.unread_by_user
            )
        else:
            # FIXME: this raises HTTP 400 and does not mark the message as read
            # https://github.com/Terrance/SkPy/issues/207
            log.debug("Calling read on %s", skype_msg)
            await self.async_wrap(skype_msg.read)

    async def correct(self, text: str, legacy_msg_id: Any, c: LegacyContact):
        pass

    async def search(self, form_values: dict[str, str]):
        pass


executor = (
    concurrent.futures.ThreadPoolExecutor()
)  # TODO: close this gracefully on exit

log = logging.getLogger(__name__)
